#ifndef __CASTLE_H__
#define __CASTLE_H__

#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include <vector>
#include <list>

#include <assert.h>

#include "globals.h"
#include "vect3.h"
#include "model.h"
#include "console.h"
#include "castle_piece.h"
#include "qsound.h"
#include "cursor.h"
#include "cannon.h"

class castle_t
{
public:
  castle_t() : mCheckEnclosureArray( 0 ), mDrawCannon( false ) {}
  ~castle_t() 
  { 
    clear(); 
    if ( mCheckEnclosureArray )
      delete [] mCheckEnclosureArray;
  }

  void reset();

  void reset_conflict_pieces();

  void clear() { mCastle.clear(); mValidPieces.clear(); }

	void draw( bool aLighting = false );
	void draw_debug();
	void update();

  void set_sound_manager( sound_manager_t *apSounds ) { mpSounds = apSounds; }

  void set_cannons( vector<cannon_t> *apCannons ) { mpCannons = apCannons; }

  void set_rect( const vect3_t &aMin, const vect3_t &aMax );
  void setFlagPos( const vect3_t &aFlagPos ) { mFlagPos = aFlagPos; }
  vect3_t getFlagPos() { return mFlagPos; }
  void setNumBoundaryPoints( const int &aNumBoundaryPoints ) { mNumBoundaryPoints = aNumBoundaryPoints; }
  void setPolyPoints( const vector<vect3_t> aPolyPoints ) { mPolyPoints = aPolyPoints; }

  const vect3_t &get_min() const { return mMin; }
  const vect3_t &get_max() const { return mMax; }

  void set_piece_model( model_node_t *apModel ) { mpPieceModel = apModel; }
  void set_piece_models( model_node_t *apModel[] );
  void set_cannon_model( model_node_t *apModel );
  void set_flag_model( model_node_t *apModels[] );
  void set_rubble_model( model_node_t *apModel[] ) 
    { mpRubbleModel[ 0 ] = apModel[ 0 ]; mpRubbleModel[ 1 ] = apModel[ 1 ]; }

  castle_piece_t *shoot();

  int get_num_valid() const { return mValidPieces.size(); }

  int get_num_cannons() const { return mpCannons->size(); }

  bool check( const vect3_t &aHitPos, const tetris_piece_t &aPiece, 
              bool aClicked );
  bool check( const vect3_t &aHitPos, bool aClicked );

  void set_tetris_piece( const tetris_piece_t &aPiece ) { mCurrPiece = aPiece; }

  // For debug only
  void remove_all_pieces();

  void enable_ghost() { mGhostEnabled = true; }
  void disable_ghost() { mGhostEnabled = false; }
  
  void enable_draw_cannon() { mDrawCannon = true; }
  void disable_draw_cannon() { mDrawCannon = false; }

  bool check_enclosure();
  void check_cannon_enclosure();
  
  int count_enclosed_pieces();

  void set_negate_sound( int aSound ) { mNegateSound = aSound; }

  void clear_rubble();

  void convert_pieces();
  void unconvert_pieces();

private:

  void set_pieces();

  bool check_piece( int aX, int aZ, const tetris_piece_t &aPiece );

  void insert_piece( int aX, int aZ, const tetris_piece_t &aPiece );

  bool piece_ok( int aX, int aZ );
  void set_piece( int aX, int aZ );
  void insert_cannon( int aX, int aZ );

  bool check_enclosure( int aX, int aY );
  void check_enclosure( int aX, int aY, bool &aEnclosed );

  void draw_flag();

  sound_manager_t *mpSounds;

  int mNumBoundaryPoints;   // Number of boundary points for the build area
  int mX, mZ;   // Number of pieces in each direction
  int mFlagBlockX, mFlagBlockZ;   // The FlagBlock is one square of the Flag area
  vector<vect3_t> mPolyPoints;     // Stores the x and z values for the boundary points
  vect3_t mMin, mMax, mFlagPos;
  vector<castle_piece_t> mCastle;
  list<int> mValidPieces; // List of indeces to valid castle pieces

  vector<cannon_t> *mpCannons;

  model_node_t *mpPieceModel;         // Model for current castle piece
  model_node_t *mpPieceModels[ 11 ];   // Model for each castle piece
  model_node_t *mpRubbleModel[ 2 ];
  model_node_t *mpCannonModel; // Model for cannon
  model_node_t *mpFlagModels[ 10 ];

  tetris_piece_t mCurrPiece;
  int mPieceX, mPieceZ; // Piece location the user points to

  bool mGhostEnabled;   // Do we draw the alpha blended tetris piece
                        // we're going to place?
  bool mDrawCannon;     // Do we draw the cannon where it's supposed to
                        // be placed?
  
  float mCannonOffX, mCannonOffZ;

  bool *mCheckEnclosureArray;

  int mNegateSound;

  struct conflict_piece_t
  {
    bool mOK;
    int mX, mZ;
  };

  conflict_piece_t mConflictPieces[ 4 ];
};

#endif