#ifndef __CONSOLE__
#define __CONSOLE__

#include <windows.h>
#include <gl\gl.h>
#include <list>
#include <vector>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string>
#include <stdarg.h>
#include <assert.h>

#include "glext.h"

#include "fmath.h"
#include "font.h"
#include "log.h"
#include "globals.h"
#include "textures.h"

using namespace std;

enum FUNC_TYPE { FT_BOOL, FT_FLOAT, FT_INT, FT_STRING, FT_COMMAND };

enum HANDLER_TYPE { HD_NORMAL, HD_GL };

enum { LINE_WIDTH = 60 };

struct console_entry_t
{
  FUNC_TYPE type;
  string name;
  void *data;
  HANDLER_TYPE hdType;  // True if the handler needs to be called after GL is 
                        // initialized
  bool constant;

  void ( *handlerf )( float );
  void ( *handlerb )( bool );
  void ( *handleri )( int );
  void ( *handlers )( const char * );
  void ( *handlercmd )( void );
};

struct Line
{
  char data[ LINE_WIDTH ];
};

class Textures;

/////////////////////////////////////////////////////////////////////
class console_t
{
public:
  console_t( int pNumLines = 12 );
  ~console_t() { entryList.clear(); delete [] lines; }

  void init( texture_manager_t *apTextures, font_t *apFont );

  void clean();
  void reset();
  
  void draw();
  void exec_file( const char *fileName );
  void save_file( const char *fileName );
  void save_options( const char *fileName );
  void reg_float( const string &name, void *data, bool constant,
		          HANDLER_TYPE handlerType, 
				  void ( *handler )( float ) = NULL );
  void reg_int( const string &name, void *data, bool constant,
		        HANDLER_TYPE handlerType, 
		        void ( *handler )( int ) = NULL );
  void reg_bool( const string &name, void *data, bool constant,
		         HANDLER_TYPE handlerType, 
		         void ( *handler )( bool ) = NULL );
  void reg_string( const string &name, void *data, bool constant,
		           HANDLER_TYPE handlerType, 
		           void ( *handler )( const char * ) = NULL );
  void reg_cmd( const string &name, void ( *handler )( void ) );

  bool set( const string &name, const char *val );
  bool toggle( const string &name );
  bool echo( const string &name );
  void add( unsigned short key );
  bool process_cmd( const string &command );
  void add_char( unsigned short key );
  void add_msg( const char *msg, ... );
  void add_instant_msg( const char *msg, ... );
  void shift();

  void set_enable() { enable = true; }
  void disable() { enable = false; }
  bool is_enabled() const { return enable; }
  void toggle();
  int get_cursor_pos() const { return cursorPos; }

  void exec_delay_handlers();

private:
  void parse_line( const char *line, bool fromFile = false );

  font_t *mpFont;
  texture_manager_t *mpTextures;
  bool enable;
  int numLines;
  list<console_entry_t> entryList;
  Line *lines;

  // List of entries that need to be run after GL is initialized
  list<console_entry_t*> mGLEntries;

  // Are we getting commands from a file?
  bool mFromFile;

  list<Line>::iterator cmdIt;
  
  int maxPos, cursorPos;

  float cursorCount;
  int conTexID;        // console_t background texture

  float tcinc;
  float drawY, yinc;

  char mInstantMsg[ 100 ];
  bool mInstantMsgValid;
  float mInstantMsgCount;
};

extern console_t *gCon;

#endif
