#ifndef __GAME_H__
#define __GAME_H__

#include <windows.h>
#include <gl\gl.h>
#include <assert.h>

#include <vector>

#include "log.h"

#include "globals.h"
#include "vect3.h"
#include "qsound.h"
#include "textures.h"
#include "modellist.h"
#include "font.h"
#include "shader.h"

#include "ship.h"
#include "cannon.h"
#include "castle.h"
#include "projectile.h"
#include "projectile2.h"
#include "particle.h"
#include "light.h"
#include "score_list.h"

class game_t
{
public:
	
	enum { EASY = 0, HARD };

	game_t();
	~game_t();

	void update();
	void draw();
	void draw_text();

  void restart();

	void load_level( const char *apFileName );

	void set_texture_manager( texture_manager_t *apTextures );
	void set_sound_manager( sound_manager_t *apSounds );
  void set_font( font_t *apFont ) { mpFont = apFont; }

  void toggle_pause() { mPause = !mPause; }

  void handle_left_mouse( bool aLeftClicked );
  void handle_right_mouse( bool aLeftClicked );

  void zoom_in();
  void zoom_out();

  void rotate_left();
  void rotate_right();

  // This is for debug only
  void clear_castle() { mCastle.remove_all_pieces(); }

private:

  enum stage_t { PLACE_CANNON = 0, PLACE_CANNON_BONUS, PRE_BATTLE, BATTLE, POST_BATTLE, 
                 PRE_BUILD, BUILD, GAME_OVER, WAIT_TO_RESTART };

  enum { PCOUNT = 20 };

	game_t( const game_t &aGame );	
	const game_t &operator=( const game_t &aGame );

  void update_battle();
  void update_postbattle();
  void update_prebattle();
  void update_prebuild();
  void update_build();
  void update_placecannon();
  void update_highscores();
  
  void draw_battle();
  void draw_build();

	void draw_text_battle();
	void draw_text_prebattle();
	void draw_text_prebuild();
	void draw_text_build();
	void draw_text_placecannon();
  void draw_text_placecannonbonus();
	void draw_text_gameover();

  void handle_mouse_battle();
  void handle_mouse_build( bool aClicked );
  void handle_mouse_placecannon( bool aClicked );

	void clear();

	void parse_ship( int &aLineCount, int shipType );
	void parse_model( int &aLineCount );
  void parse_castle( int &aLineCount );
  void parse_light( int &aLineCount );

	void draw_debug_ai_area();
  void draw_coord_ref();

  void draw_next_piece();

	float get_ship_velocity();

  void draw_select_battle();
  void draw_select_build();

  void player_shoot( const vect3_t &aTargetPos );

  bool update_player_projectiles();
  void update_ships( bool aFire );
  
  void load_constant_data();

  void reset_ships();

  projectile_t *find_free_projectile();
  projectile2_t *find_free_projectile2();

  void add_water_particles( const vect3_t &aPos );
  void add_wake_particles( ship_t &aShip, const vect3_t &aDstPos );
  void add_smoke_particles( ship_t &aShip );
  void add_cannon_particles( const vect3_t &aSrcPos, const vect3_t &aDstPos );
  void add_explosion_particles( const vect3_t &aPos );

  void draw_particles();
  void disable_particles();

  void draw_screen_quad();

  shader_manager_t mShaderManager;

	texture_manager_t *mpTextures;
	sound_manager_t *mpSounds;
  font_t *mpFont;

	bool mValid;
  stage_t mStage;

	vector<model_node_t> mWorldModels;

	vector<ship_t> mShips;
  vector<int> mOrigShipType;    // used in restart()
  int mNumValidShips;

  castle_t mCastle;

  vector<cannon_t> mCannons;

  vector<segment_t> mSegments;
  vector<vect3_t> mShipPoly;     // Polygon that defines ships' legal area

  projectile_t *mProjectiles;
  projectile2_t *mProjectiles2;
  int mNumPlayerProjs;
  int mNumShipProjs;

	model_manager_t mModelManager;

  model_node_t mCastleModel[ 11 ];  // Model for each castle piece
  model_node_t mRubbleModel[ 2 ];
  model_node_t mCannonBallModel;   // Duh!
  shadow_model_node_t *mShipModel; // Duh!
  shadow_model_node_t mShipModelLow;  
  shadow_model_node_t mShipModelHigh;  
  shadow_model_node_t *mShip2Model;
  shadow_model_node_t mShip2ModelLow;
  shadow_model_node_t mShip2ModelHigh;
  model_node_t mCannonModel;       // Duh!
  model_node_t mFlagModel[ 8 ];
  model_node_t mFlagPoleModel;
  model_node_t mFlagBaseModel;

  particle_sys_t *mWaterParticles;     // Array of particle systems
  particle_sys_t *mWakeParticles;      // Array of particle systems
  particle_sys_t *mSmokeParticles;     // Array of particle systems
  particle_sys_t *mCannonParticles;    // Array of particle systems
  particle_sys_t *mExplosionParticles; // Array of particle systems

  int mCannonSound;
  int mExplosionSound;
  int mNegateSound;

  bool mPause;

  float mRestartDelay; // How long before the user can restart a gama after a game over

  // For time
  int mSeconds;
  float mSecondCount;
  int mBuildRound;    // Used to reduce the build time
  int mBattleRound;   // Used to reduce the ship's fire delay
  int mPreviousSecond;    // Used for the build countdown sound

  GLUquadric *mQuadric;          // For debug only

  int mScore;                    // Player's score
  char mName[16];                // Player's name 
  
  vect3_t mViewerPos, mLookAtPos, mViewDir;
  float mZoom, mViewAngle;

  int mPieceTextures[ 8 ];
  tetris_piece_t mNextPiece;
  tetris_piece_t mCurrPiece;

  int mTerritoryBonus;
  float mBonusTime;

  light_t mGlobalLight;
};

#endif
