#include "model.h"

//////////////////////////////////////////////////////////////////////
model_t::model_t()
{
  mNumInds = 0;
  mpVerts = 0;
  mpTCoords = 0;
  mpNormals = 0;
  mpInds32 = 0;

  mIndType = GL_UNSIGNED_INT;
}

//////////////////////////////////////////////////////////////////////
model_t::model_t( const model_t &aModel )
{
  mName = aModel.mName;
  mNumInds = aModel.mNumInds;
  mNumVerts = aModel.mNumVerts;
  mIndType = aModel.mIndType;

  mpVerts = new vect3_t [ mNumVerts ];
  memcpy( ( vect3_t * ) mpVerts, ( vect3_t * ) aModel.mpVerts,
          mNumVerts * sizeof( vect3_t ) );

  mpTCoords = new vect2_t [ mNumVerts ];
  memcpy( ( vect3_t * ) mpTCoords, ( vect3_t * ) aModel.mpTCoords,
          mNumVerts * sizeof( vect2_t ) );

  mpNormals = new vect3_t [ mNumVerts ];
  memcpy( ( vect3_t * ) mpNormals, ( vect3_t * ) aModel.mpNormals,
          mNumVerts * sizeof( vect3_t ) );

  mpInds32 = new uint32 [ mNumInds ];
  memcpy( ( uint32 * ) mpInds32, ( uint32 * ) aModel.mpInds32,
          mNumInds * sizeof( uint32 ) );

  mSphere = aModel.mSphere;
  mMin = aModel.mMin;
  mMax = aModel.mMax;
}

//////////////////////////////////////////////////////////////////////
const model_t &
model_t::operator=( const model_t &aModel )
{
	if ( this != &aModel )
	{
    mName = aModel.mName;
    mNumInds = aModel.mNumInds;
    mNumVerts = aModel.mNumVerts;
    mIndType = aModel.mIndType;

    this->~model_t();

    mpVerts = new vect3_t [ mNumVerts ];
    memcpy( ( vect3_t * ) mpVerts, ( vect3_t * ) aModel.mpVerts,
            mNumVerts * sizeof( vect3_t ) );

    mpTCoords = new vect2_t [ mNumVerts ];
    memcpy( ( vect3_t * ) mpTCoords, ( vect3_t * ) aModel.mpTCoords,
            mNumVerts * sizeof( vect2_t ) );

    mpNormals = new vect3_t [ mNumVerts ];
    memcpy( ( vect3_t * ) mpNormals, ( vect3_t * ) aModel.mpNormals,
            mNumVerts * sizeof( vect3_t ) );

    mpInds32 = new uint32 [ mNumInds ];
    memcpy( ( uint32 * ) mpInds32, ( uint32 * ) aModel.mpInds32,
            mNumInds * sizeof( uint32 ) );

    mSphere = aModel.mSphere;
    mMin = aModel.mMin;
    mMax = aModel.mMax;
	}
	return *this;
}

//////////////////////////////////////////////////////////////////////
model_t::~model_t()
{
  if ( mpVerts )
    delete [] mpVerts;
  if ( mpTCoords )
    delete [] mpTCoords;
  if ( mpNormals )
    delete [] mpNormals;
  if ( mpInds32 )
    delete [] mpInds32;
}

//////////////////////////////////////////////////////////////////////
bool 
model_t::load( const char *aFileName, float aScale )
{
  char *lExt = COM_FileExtension( aFileName );

  if ( strcmp( lExt, "obj" ) == 0 )
    return load_text( aFileName, aScale );
  else if ( strcmp( lExt, "b3d" ) == 0 )
    return load_bin( aFileName, aScale );

  gLog->add_msg( "ERROR: Unknown file extension '%s'", lExt );
  return false;
}

//////////////////////////////////////////////////////////////////////
bool 
model_t::load_bin( const char *aFileName, float aScale )
{
  string lPath = "data/models/";
  lPath += aFileName;

  FILE *lpFile = fopen( lPath.c_str(), "rb" );

  if ( !lpFile )
  {
    gLog->add_msg( "ERROR: Could not open model '%s'", lPath.c_str() );
    return false;
  }

  mName = aFileName;

  struct b3d_header_t
  {
    uint8 identifier[ 2 ];
    uint16 num_indices;
    uint16 num_vertices;
  };

  b3d_header_t lHeader;

  fread( &lHeader, sizeof( lHeader ), 1, lpFile );
  fread( &mSphere, sizeof( sphere_t ), 1, lpFile );
  fread( &mMin, sizeof( vect3_t ), 1, lpFile );
  fread( &mMax, sizeof( vect3_t ), 1, lpFile );

  mNumInds = lHeader.num_indices;
  mNumVerts = lHeader.num_vertices;

  // Allocate memory
  mpVerts = new vect3_t [ mNumVerts ];
  mpTCoords = new vect2_t [ mNumVerts ];
  mpNormals = new vect3_t [ mNumVerts ];

  mpInds32 = new uint32 [ lHeader.num_indices ];
  fread( mpInds32, sizeof( uint32 ), lHeader.num_indices, lpFile );
  mIndType = GL_UNSIGNED_INT;

  // Now read the vertex data
  fread( mpVerts, sizeof( vect3_t ), lHeader.num_vertices, lpFile );
  fread( mpNormals, sizeof( vect3_t ), lHeader.num_vertices, lpFile );
  fread( mpTCoords, sizeof( vect2_t ), lHeader.num_vertices, lpFile );

  if ( aScale != 1.0f )
  {
    for ( int i = 0; i < mNumVerts; ++i )
    {
      mpVerts[ i ] *= aScale;
    }
  }

  gLog->add_msg( "Loaded binary model '%s'", aFileName );

  return true;
}

//////////////////////////////////////////////////////////////////////
void 
model_t::dump_gometry()
{
  int i;
  for ( i = 0; i < mNumVerts; i++ )
  {
    gLog->add_msg( "%0.2f %0.2f %0.2f : %0.2f %0.2f %0.2f : %0.2f %0.2f", 
                   mpVerts[ i ].x, mpVerts[ i ].y, mpVerts[ i ].z, 
                   mpNormals[ i ].x, mpNormals[ i ].y, mpNormals[ i ].z,
                   mpTCoords[ i ].x, mpTCoords[ i ].y ); 
  }
}

//////////////////////////////////////////////////////////////////////
bool 
model_t::load_text( const char *aFileName, float aScale )
{
  string lPath = "data/models/";
  lPath += aFileName;

  FILE *lpFile = fopen( lPath.c_str(), "r" );

  if ( !lpFile )
  {
    gLog->add_msg( "ERROR: Could not open model '%s'", lPath.c_str() );
    return false;
  }

  mName = aFileName;

  vector<index_t> lInds, lUniqInds; 
  vector<vect3_t> lVerts, lNorms;
  vector<vect2_t> lTCCoords;
  
  vect3_t lVec3, lNorm;
  vect2_t lVec2;

  char lLine[ 100 ];
  while ( fgets( lLine, 100, lpFile ) )
  {
    char *tok = strtok( lLine, " \n\t" );
    
	  if ( !tok || tok[ 0 ] == '#' )
      continue;

	  if ( strcmp( tok, "v" ) == 0 )
	  {
      tok = strtok( NULL, " \n\t" );
	    sscanf( tok, "%f", &lVec3.x );
      tok = strtok( NULL, " \n\t" );
	    sscanf( tok, "%f", &lVec3.y );
      tok = strtok( NULL, " \n\t" );
	    sscanf( tok, "%f", &lVec3.z );

      lVec3 *= aScale;

	    lVerts.push_back( lVec3 );
	  }
	  else if ( strcmp( tok, "vt" ) == 0 )
	  {
      tok = strtok( NULL, " \n\t" );
	    sscanf( tok, "%f", &lVec2.x );
      tok = strtok( NULL, " \n\t" );
	    sscanf( tok, "%f", &lVec2.y );

	    lTCCoords.push_back( lVec2 );
	  }
	  else if ( strcmp( tok, "vn" ) == 0 )
	  {
      tok = strtok( NULL, " \n\t" );
	    sscanf( tok, "%f", &lNorm.x );
      tok = strtok( NULL, " \n\t" );
  	  sscanf( tok, "%f", &lNorm.y );
      tok = strtok( NULL, " \n\t" );
      sscanf( tok, "%f", &lNorm.z );

  	  lNorm.normalize();

  	  lNorms.push_back( lNorm );
  	}
  	if ( strcmp( tok, "f" ) == 0 )
  	{
	    index_t lIndex[ 3 ];

	    tok = strtok( NULL, " \n\t" );
	    sscanf( tok, "%d/%d/%d", &lIndex[ 0 ].i0, &lIndex[ 0 ].i1, 
		          &lIndex[ 0 ].i2 );
	    lIndex[ 0 ].i0--;
	    lIndex[ 0 ].i1--; 
	    lIndex[ 0 ].i2--;

	    tok = strtok( NULL, " \n\t" );
	    sscanf( tok, "%d/%d/%d", &lIndex[ 1 ].i0, &lIndex[ 1 ].i1, 
		          &lIndex[ 1 ].i2 );
      lIndex[ 1 ].i0--;
      lIndex[ 1 ].i1--;
	    lIndex[ 1 ].i2--;

	    tok = strtok( NULL, " \n\t" );
	    sscanf( tok, "%d/%d/%d", &lIndex[ 2 ].i0, &lIndex[ 2 ].i1, 
		          &lIndex[ 2 ].i2 );
      lIndex[ 2 ].i0--;
      lIndex[ 2 ].i1--;
	    lIndex[ 2 ].i2--;

      lInds.push_back( lIndex[ 0 ] );
      lInds.push_back( lIndex[ 1 ] );
      lInds.push_back( lIndex[ 2 ] );

	    check_index( lUniqInds, lIndex[ 0 ] );
	    check_index( lUniqInds, lIndex[ 1 ] );
	    check_index( lUniqInds, lIndex[ 2 ] );
	  } 
  }

  mNumInds = lInds.size();
  mIndType = GL_UNSIGNED_INT;

  // Now build final arrays
  mNumVerts = lUniqInds.size();
  mpVerts = new vect3_t [ mNumVerts ];
  mpTCoords = new vect2_t [ mNumVerts ];
  mpNormals = new vect3_t [ mNumVerts ];

  mpInds32 = new uint32 [ mNumInds ];

  int i, j;
  for ( i = 0; i < mNumVerts; i++ )
  {
	  index_t ind = lUniqInds[ i ];

    mpVerts[ i ] = lVerts[ ind.i0 ];
	
    mpTCoords[ i ] = lTCCoords[ ind.i1 ];

    mpNormals[ i ] = lNorms[ ind.i2 ];
  }

  for ( i = 0; i < lInds.size(); i++ )
  {
    for ( j = 0; j < lUniqInds.size(); j++ )
    {
      if ( lInds[ i ].i0 == lUniqInds[ j ].i0 &&
		       lInds[ i ].i1 == lUniqInds[ j ].i1 &&
		       lInds[ i ].i2 == lUniqInds[ j ].i2 )
	    {
        mpInds32[ i ] = j;
	    }
	  }
  }

  //gen_vertex_normals( lInds, size );
  //gen_face_normals();

  gen_sphere();
  gen_box();

  lUniqInds.clear();
  lInds.clear();
  lVerts.clear();
  lTCCoords.clear();

  fclose( lpFile );
  
  gLog->add_msg( "Loaded model '%s'", aFileName );

  return true;
}

//////////////////////////////////////////////////////////////////////
// Make sure the array only has one instance of each element
//////////////////////////////////////////////////////////////////////
void 
model_t::check_index( vector<index_t> &aIndeces, const index_t aIndex )
{
	for ( int i = 0; i < aIndeces.size(); i++ )
	{
		if ( aIndex.i0 == aIndeces[ i ].i0 && 
			   aIndex.i1 == aIndeces[ i ].i1 && 
			   aIndex.i1 == aIndeces[ i ].i2 )
			return;
	}

	aIndeces.push_back( aIndex );
}

//////////////////////////////////////////////////////////////////////
void 
model_t::draw()
{
  glVertexPointer( 3, GL_FLOAT, 0, mpVerts );
  glNormalPointer( GL_FLOAT, 0, mpNormals );
  glTexCoordPointer( 2, GL_FLOAT, 0, mpTCoords );

  glLockArrays( 0, mNumVerts );
  glDrawElements( GL_TRIANGLES, mNumInds, mIndType, mpInds32 );
  glUnlockArrays();
}

//////////////////////////////////////////////////////////////////////
void 
model_t::set_pointers()
{
  glVertexPointer( 3, GL_FLOAT, 0, mpVerts );
  glNormalPointer( GL_FLOAT, 0, mpNormals );
  glTexCoordPointer( 2, GL_FLOAT, 0, mpTCoords );
  glLockArrays( 0, mNumVerts );
}

//////////////////////////////////////////////////////////////////////
void 
model_t::unset_pointers()
{
  glUnlockArrays();
}

//////////////////////////////////////////////////////////////////////
void 
model_t::draw_pointers()
{
  glDrawElements( GL_TRIANGLES, mNumInds, mIndType, mpInds32 );
}

//////////////////////////////////////////////////////////////////////
void 
model_t::draw_lines()
{
//  glVertexPointer( 3, GL_FLOAT, 0, mpVerts );
//  glNormalPointer( GL_FLOAT, 0, mpNormals );
//  glTexCoordPointer( 2, GL_FLOAT, 0, mpTCoords );

  glDrawElements( GL_LINES, mNumInds, mIndType, mpInds32 );
}

//////////////////////////////////////////////////////////////////////
void 
model_t::gen_face_normals()
{
	static vect3_t v0, v1, v2, vect1, vect2, normal;
	static int i;

	for ( i = 0; i < mNumInds / 3; i++ )
	{
		int a = i * 3;
	
		int offset0 = mpInds32[ a + 0 ];
		int offset1 = mpInds32[ a + 1 ];
		int offset2 = mpInds32[ a + 2 ];

		v0 = mpVerts[ offset0 ];
		v1 = mpVerts[ offset1 ];
		v2 = mpVerts[ offset2 ];
  
		// Generate two vectors on this plane
		vect1 = v1 - v0;
		vect2 = v2 - v1;

		normal = vect1.cross( vect2 );
		normal.normalize();

		mpNormals[ offset0 ] = normal;
		mpNormals[ offset1 ] = normal;
		mpNormals[ offset2 ] = normal;
	}   
}

//////////////////////////////////////////////////////////////////////
// Draw normals for debug
//////////////////////////////////////////////////////////////////////
void
model_t::draw_normals()
{
	glColor3f( 1, 0, 0 );
	glBegin( GL_LINES );

	for ( int i = 0; i < mNumInds; i++ )
	{
		vect3_t n = mpNormals[ i ];

		vect3_t v = mpVerts[ i ];

		vect3_t v2 = v + n;
		
		glVertex3f( v.x, v.y, v.z );
		glVertex3f( v2.x, v2.y, v2.z );
	}
	glEnd();
}

//////////////////////////////////////////////////////////////////////
// Generate bounding box
//////////////////////////////////////////////////////////////////////
void 
model_t::gen_box()
{
	mMin = 1e22f;
	mMax = 1e-22f;
	vect3_t v;

	for ( int i = 0; i < mNumVerts; i++ )
	{
		v = mpVerts[ i ];

		if ( v.x > mMax.x )
			mMax.x = v.x;

		if ( v.y > mMax.y )
			mMax.y = v.y;

		if ( v.z > mMax.z )
			mMax.z = v.z;

		if ( v.x < mMin.x )
			mMin.x = v.x;

		if ( v.y < mMin.y )
			mMin.y = v.y;

		if ( v.z < mMin.z )
			mMin.z = v.z;
	}
}

//////////////////////////////////////////////////////////////////////
// Generate bounding sphere
//////////////////////////////////////////////////////////////////////
void 
model_t::gen_sphere()
{
	mSphere.mCenter = 0.0;

	int i;
	for ( i = 0; i < mNumVerts; i++ )
	{
		mSphere.mCenter += mpVerts[ i ];  
	}
  
	mSphere.mCenter /= ( float ) mNumVerts;
	mSphere.mRad.f = 0;

	vect3_t v;
	for ( i = 0; i < mNumVerts; i++ )
	{
		v = mpVerts[ i ];

		float dis = v.distance( mSphere.mCenter );
    
		if ( dis > mSphere.mRad.f )
			mSphere.set_rad( dis );
	}

  mSphere.square_rad();
}

////////////////////////////////////////////////////////////////////////
model_node_t::model_node_t() 
{
	mpShader = NULL;
	mpModel = NULL;
}

////////////////////////////////////////////////////////////////////////
void 
model_node_t::set_shader( shader_t *apShader )
{	
	mpShader = apShader;
	assert( mpShader );

	for ( int i = 0; i < mpShader->passes.size(); i++ )
	{
		PolyPassData data;

		for ( int tcpass = 0; tcpass < mpShader->passes[ i ].tcData.size(); tcpass++ )
		{
			ShaderPassData spd;
			data.tcData.push_back( spd );
			data.tcFlags.push_back( mpShader->passes[ i ].tcFlags[ i ] );
		}

		mShaderData.push_back( data );
	}

}

////////////////////////////////////////////////////////////////////////
void 
model_node_t::draw()
{
	if ( mpShader )
	{
		draw_shader();
	}
	else
	{
		glBindTexture( GL_TEXTURE_2D, mTexture );
		mpModel->draw();
	}
}

////////////////////////////////////////////////////////////////////////
void 
model_node_t::draw_pointers()
{
	if ( mpShader )
	{
		draw_shader();
	}
	else
	{
		glBindTexture( GL_TEXTURE_2D, mTexture );
		mpModel->draw_pointers();
	}
}
 
////////////////////////////////////////////////////////////////////////
void 
model_node_t::draw_lines()
{
//	if ( mpShader )
//	{
//		draw_shader();
//	}
//	else
//	{
//		glBindTexture( GL_TEXTURE_2D, mTexture );
		mpModel->draw_lines();
//	}
}
 
////////////////////////////////////////////////////////////////////////
void 
model_node_t::draw_pos()
{
  glPushMatrix(); 
  glTranslatef( mPos.x, mPos.y, mPos.z ); 
  draw();
  glPopMatrix();
}

////////////////////////////////////////////////////////////////////////
void 
model_node_t::draw_shader()
{	
	static ShaderPass *pass;
	static PolyPassData *data;

	if ( mpShader->mFlags & SF_INVISIBLE )
		return;

  // Enable blending for rendering geometry with shaders
  glEnable( GL_BLEND );

	if ( mpShader->mFlags & SF_DEPTH_OFF )
		glDepthMask( GL_FALSE );

	if ( mpShader->mFlags & SF_NO_CULL )
		glDisable( GL_CULL_FACE );

  if ( mpShader->mFlags & SF_TRANSFORM )
  {
    transform_vertices();
  }

	// Now we draw each pass
	for ( int i = 0; i < mpShader->passes.size(); i++ )
	{
		// If we have multiple passes -> set depth test to equal
		if ( i == 1 )			
			glDepthFunc( GL_EQUAL );

		pass = &mpShader->passes[ i ];
		data = &( mShaderData[ i ] );

		// Update the current shader pass
		update_shader_pass( i );

		// Draw the model
		mpModel->draw();

		if ( pass->mFlags & SF_TCMOD )
		{
			glMatrixMode( GL_TEXTURE );
			glLoadIdentity();
			glMatrixMode( GL_MODELVIEW );
		}

		if ( pass->mFlags & SF_ALPHA )
		{
			glDisable( GL_ALPHA_TEST );
		}   

		if ( pass->mFlags & SF_TCGEN )
		{
			glEnableClientState( GL_NORMAL_ARRAY );
			glDisable( GL_TEXTURE_GEN_S );
			glDisable( GL_TEXTURE_GEN_T );
		}	  
	}

	glDepthMask( GL_TRUE );
	glEnable( GL_CULL_FACE );
	glDepthFunc( GL_LESS );
  glDisable( GL_BLEND );
}

////////////////////////////////////////////////////////////////////////
void 
model_node_t::update_shader_pass( int aNum )
{	
	static ShaderPass *pass;
	static PolyPassData *data;

	pass = &mpShader->passes[ aNum ];
	data = &( mShaderData[ aNum ] );

	data->mCurrTexture = pass->mTextures[ 0 ];

	if ( pass->mFlags & SF_TEXCYCLE )
	{
		data->mCurrTexture = pass->mTextures[ data->cycleData.d1.iVal ];
		if ( data->cycleData.d0.iVal++ == pass->cycleData.d1.iVal )
		{
			data->cycleData.d0.iVal = 0;
			if ( ++data->cycleData.d1.iVal == pass->cycleData.d1.iVal )
				data->cycleData.d1.iVal = 0;
		}    
	}  
	
	glBindTexture( GL_TEXTURE_2D, data->mCurrTexture );

	if ( pass->mFlags & SF_ALPHA )
	{
		glEnable( GL_ALPHA_TEST );
	}	  

	if ( pass->mFlags & SF_TCGEN )
	{
		glDisableClientState( GL_NORMAL_ARRAY );
		glTexGeni( GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP );
		glTexGeni( GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP );
		glEnable( GL_TEXTURE_GEN_S );
		glEnable( GL_TEXTURE_GEN_T );
	}	

	if ( pass->mFlags & SF_RGBGEN )
	{
		if ( pass->mFlags & SF_SQUARE )
		{
			// X++
			data->rgbData.d0.fVal += gFactTick;

			if ( data->rgbData.d0.fVal < pass->rgbData.d2.fVal )
			{
				data->rgbData.d1.fVal = 0; // Y = 0
			}
			else
			{
				data->rgbData.d1.fVal = pass->rgbData.d1.fVal; // Y = YMax
			}

			if ( data->rgbData.d0.fVal > ( 2 * pass->rgbData.d2.fVal ) )
			{
				data->rgbData.d0.fVal = 0;
			}

			// Final RGB = currY + YMin
			pass->mRGB = data->rgbData.d1.fVal + pass->rgbData.d0.fVal;
		}
		else if ( pass->mFlags & SF_TRIANGLE )
		{
			data->rgbData.d0.fVal += gFactTick;

			if ( data->rgbData.d0.fVal < pass->rgbData.d2.fVal )
			{
				data->rgbData.d1.fVal += pass->rgbData.d1.fVal * gFactTick;
			}
			else
			{
				data->rgbData.d1.fVal -= pass->rgbData.d1.fVal * gFactTick;
			}

			if ( data->rgbData.d0.fVal > ( 2 * pass->rgbData.d2.fVal ) )
			{
				data->rgbData.d0.fVal = 0;
				data->rgbData.d1.fVal = 0;
			}

			// Final RGB = currY + YMin
			pass->mRGB = data->rgbData.d1.fVal + pass->rgbData.d0.fVal;
		}
		else if ( pass->mFlags & SF_SAWTOOTH )
		{
			data->rgbData.d0.fVal += gFactTick;

			if ( data->rgbData.d0.fVal < pass->rgbData.d2.fVal )
			{
				data->rgbData.d1.fVal += pass->rgbData.d1.fVal * gFactTick;
			}
			else
			{
				data->rgbData.d0.fVal = 0;
				data->rgbData.d1.fVal = 0;
			}

			// Final RGB = currY + YMin
			pass->mRGB = data->rgbData.d1.fVal + pass->rgbData.d0.fVal;
		}
		else if ( pass->mFlags & SF_SINE )
		{
			data->rgbData.d0.fVal += gFactTick * pass->rgbData.d3.fVal;
			data->rgbData.d1.fVal = fsin( data->rgbData.d0.fVal ) * pass->rgbData.d1.fVal;

			// Final RGB = currY + YMin
			pass->mRGB = data->rgbData.d1.fVal + pass->rgbData.d0.fVal;
		}
	}

	if ( pass->mFlags & SF_TCMOD )
	{
		glMatrixMode( GL_TEXTURE );

		for ( int tcpass = 0; tcpass < pass->tcData.size(); tcpass++ )
		{
			unsigned int lFlag = pass->tcFlags[ tcpass ];
			ShaderPassData *shData = &pass->tcData[ tcpass ];	
			ShaderPassData *paData = &data->tcData[ tcpass ];		

			switch ( lFlag )
			{
			case SF_SCROLL :
				{
					paData->d0.fVal += shData->d0.fVal * gFactTick;
					paData->d1.fVal += shData->d1.fVal * gFactTick;
					glTranslatef( paData->d0.fVal, paData->d1.fVal, 0 );
					break;
				}
			case SF_ROTATE :
				{
					paData->d0.fVal += shData->d0.fVal * gFactTick;
					glTranslatef( 0.5, 0.5, 0 );
					glRotatef( paData->d0.fVal, 0.0, 0.0, 1.0 );
					glTranslatef( -0.5, -0.5, 0 );
					break;
				}
			case SF_SCALE :
				{
					if ( paData->d1.iVal == 0 )
					{
						paData->d0.fVal += shData->d0.fVal * gFactTick;
						if ( paData->d0.fVal >= shData->d2.fVal )
						{
							paData->d0.fVal = shData->d2.fVal;
							paData->d1.iVal = !paData->d1.iVal;
						}
					}
					else
					{
						paData->d0.fVal -= shData->d0.fVal * gFactTick;
						if ( paData->d0.fVal < shData->d1.fVal )
						{
							paData->d0.fVal = shData->d1.fVal;
							paData->d1.iVal = !paData->d1.iVal;
						}
					}
					glTranslatef( 0.5, 0.5, 0 );
					glScalef( paData->d0.fVal, paData->d0.fVal, 0 );
					glTranslatef( -0.5, -0.5, 0 );
					break;
				}		
			}
		}
		glMatrixMode( GL_MODELVIEW );
	}

	glColor4f( pass->mRGB, pass->mRGB, pass->mRGB, pass->mRGB );

	glBlendFunc( pass->mSrc, pass->mDst );
}
 
////////////////////////////////////////////////////////////////////////
void 
model_node_t::transform_vertices()
{
  static float lAnim = 0, lFreq, lAmp;
  static vect3_t *lpVect;

  lFreq = mpShader->mXferFunction.mArgs[ 0 ];
  lAmp = mpShader->mXferFunction.mArgs[ 1 ];

  lAnim -= lFreq * gFactTick;

  for ( int i = 0; i < mpModel->get_num_vertices(); ++i )
  {
    lpVect = &( mpModel->get_vertices()[ i ] );

    //y = sin( ( ( x + z ) * freq ) + anim ) * amp;
    //lpVect->y = sin( ( ( lpVect->x + lpVect->z ) * lFreq ) + lAnim ) * lAmp;
    lpVect->y = sin( ( ( ( lpVect->x + lpVect->z ) *100) + lAnim ) ) * lAmp;
  }
}
