#ifndef __SHADOW_MODEL__
#define __SHADOW_MODEL__

#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <string>
#include <list>
#include <vector>
#include <memory.h>

#include "globals.h"
#include "glext.h"
#include "ijl.h"
#include "misc.h"
#include "log.h"
#include "vect3.h"
#include "shader.h"
#include "model.h"

struct tri_t
{
  int inds[ 3 ]; // Indices into 3 adjacent triangles
  bool front_facing;
};

////////////////////////////////////////////////////////////////////////
class shadow_model_t
{
public :
  shadow_model_t();
  virtual ~shadow_model_t();

	shadow_model_t( const shadow_model_t &shadow_model_t );
	const shadow_model_t &operator=( const shadow_model_t &aModel );

  bool load( const char *aFileName, float aScale = 1 );
  
  void draw();
  void draw_lines();
  void draw_normals();

  // Shadow volume interface
  void calc_shadow_volume( const vect3_t &aLightPos );
  void draw_shadow_volume( const vect3_t &aLightPos );

  const string &get_name() const { return mName; }

  const sphere_t &get_sphere() const { return mSphere; }
  void get_box( vect3_t &aMin, vect3_t &aMax ) const { aMin = mMin; aMax = mMax; }

  vect3_t *get_vertices() { return mpVerts; }
  int get_num_vertices() const { return mNumVerts; }

protected :

  bool load_text( const char *aFileName, float aScale );
  bool load_bin( const char *aFileName, float aScale );

  void check_index( vector<index_t> &aIndeces, const index_t aIndex );
  void gen_face_normals();

  void build_shadow_indices( const vect3_t &aLightPos );

  void find_edge_connectivity();
  int find_neighbor( int aSelf, int i0, int i1 );

  void gen_sphere();
  void gen_box();

  void dump_gometry();

  string mName;

  int mNumInds; 
  int mNumFaces;
  int mNumVerts;

  GLenum mIndType;

  vect3_t *mpVerts;
  vect2_t *mpTCoords;
  vect3_t *mpNormals;
  uint32 *mpInds32; // indices

  uint32 *mpShadowInds; // Shadow volume indices

  tri_t *mpTris;

  sphere_t mSphere;
  vect3_t mMin, mMax; // Bounding box

  int mNumValidShadowInds;
};

////////////////////////////////////////////////////////////////////////
class shadow_model_node_t
{
public:
  shadow_model_node_t(); 
	~shadow_model_node_t() {}

	void set_model( shadow_model_t *apModel ) { mpModel = apModel; }
	void set_shader( shader_t *apShader );
	void set_texture( int aTexture ) { mTexture = aTexture; }
  void set_pos( const vect3_t &lPos ) { mPos = lPos; }
  const vect3_t &get_pos() const { return mPos; }

	void draw();
	void draw_lines();
  void draw_pos();

  const sphere_t &get_sphere() const { return mpModel->get_sphere(); }
  void get_box( vect3_t &aMin, vect3_t &aMax ) const 
    { mpModel->get_box( aMin, aMax ); }

  // Shadow volume interface
  void calc_shadow_volume( const vect3_t &aLightPos )
  { mpModel->calc_shadow_volume( aLightPos ); }
  void draw_shadow_volume( const vect3_t &aLightPos )
  { mpModel->draw_shadow_volume( aLightPos ); }

private:

  // Shader Stuff
	void draw_shader();
	void update_shader_pass( int aNum );
  void transform_vertices();
  //

	shadow_model_t *mpModel;

	int mTexture;

	shader_t *mpShader;

	// Data for each shader pass
	vector<PolyPassData> mShaderData;

  vect3_t mPos;
};

#endif
