#include "ship.h"

//////////////////////////////////////////////////
void 
ship_t::reset()
{
	mTurnCount = 0;

  static const float lStartVals[] = { 200, 250, 300, 350, 400, 450, 500, 550 };

  int val = rand() & 7;

  // Randomize so that the ships don't all fire at the sime time
  mShootCount = lStartVals[ val ];
	mState = STRAIGHT;

  if ( mShipType == 2 )
    mHealth = con_ship_health * 2;
  else
    mHealth = con_ship_health;

  mPos = mOrigPos;
  mDir = mOrigDir;
	mAngle = mOrigAngle;
  mVel = mOrigVel;

  mPredictPos = mPos + mDir * 100 * mVel;

  mpSmokeParticles = 0;
  mpWakeParticles = 0;

  mCollidingWithShip = false;
}

//////////////////////////////////////////////////
// Returns true if dead
//////////////////////////////////////////////////
bool 
ship_t::hit()
{
  --mHealth;
  if ( mHealth == 0 )
  {
    mState = DIEING;
    return true;
  }

  return false;
}

//////////////////////////////////////////////////
void 
ship_t::set_dir( const vect3_t &aDir ) 
{
	vect3_t lXAxis( 1, 0, 0 );

	mOrigDir = aDir;
	mOrigDir.normalize();

	// Figure out angle between direction and th X axis
	mOrigAngle = mOrigDir.angle( lXAxis ) / PI_OVER_180;
}

//////////////////////////////////////////////////
bool
ship_t::update( const vector<segment_t> &aSegments,
                const vector<vect3_t> &aShipPoly,
                vector<ship_t> &aShips,
                const int aBattleRound )
{
  static bool lCollide;
  static float lAvoidAngle, lAngleIncTick;
  static vect3_t lMin, lMax;
  static vect3_t lTailPos;

  if ( mState == DEAD || !con_ai )
    return false;

  // Move in the current direction
  mPos += mDir * gFactTick * mVel;

  // Update the smoke particles' position
  if ( mpSmokeParticles )
    mpSmokeParticles->set_pos( mPos );

  // Calculate the wake particles' position
  if ( mShipType == 1 )
  {
    lTailPos = mPos - (mDir * 20.0f);
    lTailPos.y += 5;
  }
  else
  {
    lTailPos = mPos - (mDir * 30.0f);
    lTailPos.y += 5;
  }

  // Update the wake particles' position
  if ( mpWakeParticles )
    mpWakeParticles->set_pos( lTailPos );

	switch ( mState )
	{
  case DIEING :
    {
      // Make ship sink
      mPos.y -= gFactTick * 0.1f;

      // Slow this puppy down
      mVel -= gFactTick * 0.001f;
      if ( mVel <= 0 )
        mVel = 0;

      mpModel->get_box( lMin, lMax );

      if ( mPos.y <= -lMax.y )
      {
        mState = DEAD;
        mpSmokeParticles->disable_loops();
        mpWakeParticles->disable_loops();
      }
    }
    return false;

	case STRAIGHT :
		{
      lCollide = check_collision( aSegments, aShipPoly, aShips );

			mTurnCount += gFactTick;
      if ( lCollide )
      {
				mTurnCount = 0;

				mState = AVOIDING;
      }
			else if ( mTurnCount > 300.0f )
			{
				mTurnCount = 0;

				mTargetAngle = mAngle + 90.0f * crand();

				mAngleInc = ( mTargetAngle > mAngle ) ? 0.5f : -0.5f;

        mTurnTime = 0;

				mState = TURNING;
			}			
		}
		break;

	case TURNING :
		{
      lCollide = check_collision( aSegments, aShipPoly, aShips );

      mTurnTime += gFactTick;
      if ( mTurnTime > 500.0f && !lCollide )
        mState = STRAIGHT;

      if ( lCollide )
      {
				mTurnCount = 0;

				mState = AVOIDING;
      }
      else
      {
        lAngleIncTick = mAngleInc * gFactTick;

			  mAngle += lAngleIncTick;
  			mDir.rotateY( deg2rad( lAngleIncTick ) );
			  mDir.normalize();

        // Keep the angle between 0 and 360
        if ( mAngle >= 360.0f )
          mAngle -= 360.0f;
        else if ( mAngle < -360.0f )
          mAngle += 360.0f;

			  if ( mAngleInc > 0 )
			  {
  				if ( mAngle >= mTargetAngle )
				  {
  					mState = STRAIGHT;
				  }
			  }
			  else if ( mAngleInc < 0 )
			  {
  				if ( mAngle <= mTargetAngle )
				  {
					  mState = STRAIGHT;
  				}		  		
			  }
      }
		}
		break;

  case AVOIDING :
		{
      lAngleIncTick = mAngleInc * gFactTick;

			mAngle += lAngleIncTick;
			mDir.rotateY( deg2rad( lAngleIncTick ) );
			mDir.normalize();

      // Keep the angle between 0 and 360
      if ( mAngle >= 360.0f )
        mAngle -= 360.0f;
      else if ( mAngle < -360.0f )
        mAngle += 360.0f;

      if ( !check_collision( aSegments, aShipPoly, aShips, 100 ) )
        mState = STRAIGHT;
    }
		break;
  }	

  mShootCount += gFactTick;
  static float lCount;
  if ( mShipType == 2 )
  {    
    if ( con_difficulty == HARD )
      lCount = ( con_ship_shoot_count - ( aBattleRound * 10 ) ) * 0.5f;
    else
      lCount = con_ship_shoot_count * 0.5f;
  }
  else
  {
    if ( con_difficulty == HARD )
      lCount = ( con_ship_shoot_count - ( aBattleRound * 10 ) );
    else
      lCount = con_ship_shoot_count;
  }

  if ( mShootCount > max( 200.0f, lCount ) )
  {
    mShootCount = 0;
    return true;
  }

  return false;
}

//////////////////////////////////////////////////
bool
ship_t::check_collision( const vector<segment_t> &aSegments,
                         const vector<vect3_t> &aShipPoly,
                         vector<ship_t> &aShips, float aMoveTime )
{
  static float lDistance;
  static int i;

  // See if we're on a course for collision
  // mPredictPos is only used by draw_debug
  if ( con_draw_debug )
  {
    mPredictPos = mPos;
    mPredictPos += mDir * aMoveTime * mVel;
  }

  // Check for collision with all the segments
  for ( i = 0; i < aSegments.size(); ++i )
	{
    const segment_t &lS = aSegments[ i ];

    lDistance = aSegments[ i ].dist_from_line_seg( mPos );

		if ( lDistance < 80 )
    {
      return true;
    }
  }

  if ( !mCollidingWithShip )
  {
    // Check for collision with the other ships
    for ( i = 0; i < aShips.size(); ++i )
    {
      ship_t &lShip = aShips[ i ];

      // Don't check with itself
      if ( mID == lShip.get_id() || lShip.is_alive() == false )
        continue;

      if ( mPos.distanceSquared( lShip.get_pos() ) < 20000 )
      {
        if ( lShip.get_angleinc() == 0.5f )
          mAngleInc = -0.5f;
        else
          mAngleInc = 0.5f;

        return true;
      }
    }
  }

  return false;
}

//////////////////////////////////////////////////
void
ship_t::draw()
{	
  if ( mState == DEAD )
    return;

	glPushMatrix(); 
	glTranslatef( mPos.x, mPos.y, mPos.z );  
	glRotatef( -mAngle, 0, 1, 0 ); 

	mpModel->draw();

	glPopMatrix();
}

//////////////////////////////////////////////////
void 
ship_t::draw_shadow( const vect3_t &aLightPos )
{ 
  static vect3_t lYAxis;
  lYAxis.x = 0;
  lYAxis.y = 1;
  lYAxis.z = 0;

  // Create obj->world matrix
  static mat4_t lModel;
  lModel.set_id();
  lModel.set_rot( deg2rad( -mAngle ), lYAxis );
  lModel.set_trans( mPos );

  // Create inverse model matrix
  static mat4_t lInv;
  invert_fast( lInv, lModel );  

  // Find position of light in model local space
  static vect3_t lLocalLight;    
  mult( lLocalLight, lInv, aLightPos );

  mpModel->calc_shadow_volume( lLocalLight );

  glPushMatrix();
  glMultMatrixf( lModel.mData );
  mpModel->draw_shadow_volume( lLocalLight );
  glPopMatrix();
}

//////////////////////////////////////////////////
void 
ship_t::draw_shadow_two_pass( const vect3_t &aLightPos )
{ 
  static vect3_t lYAxis;
  lYAxis.x = 0;
  lYAxis.y = 1;
  lYAxis.z = 0;

  // Create obj->world matrix
  static mat4_t lModel;
  lModel.set_id();
  lModel.set_rot( deg2rad( -mAngle ), lYAxis );
  lModel.set_trans( mPos );

  // Create inverse model matrix
  static mat4_t lInv;
  invert_fast( lInv, lModel );  

  // Find position of light in model local space
  static vect3_t lLocalLight;    
  mult( lLocalLight, lInv, aLightPos );

  mpModel->calc_shadow_volume( lLocalLight );

  glPushMatrix();
  glMultMatrixf( lModel.mData );

  // First pass
  glFrontFace( GL_CCW );
  glStencilOp( GL_KEEP, GL_KEEP, GL_INCR );
  mpModel->draw_shadow_volume( lLocalLight );

  // Second pass
  glFrontFace( GL_CW );
  glStencilOp( GL_KEEP, GL_KEEP, GL_DECR );
  mpModel->draw_shadow_volume( lLocalLight );
  
  glPopMatrix();
}

//////////////////////////////////////////////////
void
ship_t::draw_debug()
{	
  if ( mState == DEAD )
    return;
  
  glColor3f( 1, 1, 1 );
  glBegin( GL_LINES );
	glVertex3f( mPos.x, 13, mPos.z );
  glVertex3f( mPredictPos.x, 13, mPredictPos.z );
  glEnd();
}
