#include "textures.h"

//////////////////////////////////////////////////////////////////////
void
texture_manager_t::clear() 
{  
  list<CTexture>::iterator ite = tList.begin();
  while ( ite != tList.end() )
  {
    CTexture *text = &( *ite );
    glDeleteTextures( 1, &text->number );
    ite++;
  }
  tList.clear();
}

//////////////////////////////////////////////////////////////////////
int
texture_manager_t::insert( const string &texName, unsigned flags )
{
  string tName = texName;
  for ( int i = 0; i < tName.size(); i++ )
	tName[ i ] = tolower( tName[ i ] );

  list<CTexture>::iterator ite = tList.begin();
  while ( ite != tList.end() )
  {
    CTexture *text = &( *ite );
    if ( text->name == tName )
      return text->number;
    ite++;
  }

  string fileName = "data/textures/";
  fileName += tName;

  CTexture iText;
  if ( strstr( tName.c_str(), ".bmp" ) )
  {
    if ( !loadBMP( fileName.c_str(), &iText ) )
	{
      return -1;
	}
  }
  else if ( strstr( tName.c_str(), ".tga" ) )
  {
    if ( !loadTGA( fileName.c_str(), &iText ) )
	{
      return -1;
	}
  }
  else if ( strstr( tName.c_str(), ".jpg" ) )
  {
    if ( !loadJPG( fileName.c_str(), &iText ) )
	{
      return -1;
	}
  }

  unsigned int currNumber;
  glGenTextures( 1, &currNumber );
  glBindTexture( GL_TEXTURE_2D, currNumber );

  if ( flags & CLAMP )
  {
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_SGIS );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_SGIS );
  }
  else
  {
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
  }
  
  GLenum minFilter;
  if ( mFiltering == TRILINEAR )
    minFilter = GL_LINEAR_MIPMAP_LINEAR;     // Trilinear
  else
    minFilter = GL_LINEAR_MIPMAP_NEAREST;    // Bilinear

  GLenum maxFilter = GL_LINEAR;
  if ( flags & NEAREST )
  {
    minFilter = GL_NEAREST;
    maxFilter = GL_NEAREST;
  }

  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter );
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, maxFilter );

  GLenum format;
  switch ( iText.components )
  {
  case 1 :
    format = GL_ALPHA;
    break;
  case 3 :
    //format = GL_RGB;
    format = GL_BGR_EXT;
    break;
  case 4 :
    //format = GL_RGBA;
    format = GL_BGRA_EXT;
    break;
  default :
	  gLog->add_msg( "Bad texture format in %s", tName.c_str() );
  }

  gLog->add_msg( "Loading texture %s id = %d", tName.c_str(), currNumber );

//  glTexImage2D( GL_TEXTURE_2D, 0, iText.components,
//				iText.sizeX, iText.sizeY,
//				0, format, GL_UNSIGNED_BYTE, iText.data );
  
  gluBuild2DMipmaps( GL_TEXTURE_2D, iText.components, iText.sizeX, 
    iText.sizeY, format, GL_UNSIGNED_BYTE, iText.data);
  
  free( iText.data );

  iText.name = tName;
  iText.number = currNumber;
  tList.push_back( iText );
  return iText.number;
}

//////////////////////////////////////////////////////////////////////
texture_manager_t::~texture_manager_t()
{
  list<CTexture>::iterator ite = tList.begin();
  while ( ite != tList.end() )
  {
    CTexture *text = &( *ite );
    glDeleteTextures( 1, &text->number );
    ite++;
  }
  tList.clear();
}

//////////////////////////////////////////////////////////////////////
bool
texture_manager_t::loadBMP( const char *fileName, CTexture *tex )
{
  FILE *file;
  
  // make sure the file is there.
  if ( ( file = fopen( fileName, "rb" ) ) == NULL ) 
  {
    gLog->add_msg( "File %s not Found", fileName );
    return NULL;
  }

  return loadBMP( file, tex );
}

//////////////////////////////////////////////////////////////////////
bool
texture_manager_t::loadBMP( FILE *file, CTexture *tex )
{
  unsigned long size;                 // size of the image in bytes.
  unsigned long i;                    // standard counter.
  unsigned short int planes;          // number of planes in image (must be 1) 
  unsigned short int bpp;             // number of bits per pixel (must be 24)
  //unsigned char temp;                 // used to convert bgr to rgb color.

  // seek through the bmp header, up to the width/height:
  fseek( file, 18, SEEK_CUR );
  
  tex->sizeX = getint( file );
  tex->sizeY = getint( file );

  // read the planes
  planes = getshort( file );
  if ( planes != 1 ) 
  {
    //printf( "Planes from %s is not 1: %u\n", fileName, planes );
    return 0;
  }
  
  // read the bpp
  bpp = getshort( file );
  tex->components = bpp / 8;
  size = tex->sizeX * tex->sizeY * tex->components;
  
  // seek past the rest of the bitmap header.
  fseek( file, 24, SEEK_CUR );

  // read the data. 
  tex->data = ( unsigned char * ) malloc( size );
  if ( tex->data == NULL ) 
  {
    gLog->add_msg( "Error allocating memory for color-corrected image data" );
    return 0;	
  }

  if ( ( i = fread( tex->data, size, 1, file ) ) != 1 ) 
  {
    //printf( "Error reading image data from %s.\n", fileName );
    return 0;
  }
  return true;
}

//////////////////////////////////////////////////////
bool
texture_manager_t::loadTGA( const char *fileName, CTexture *tex )
{
  FILE *file;
  unsigned long size;                 // size of the image in bytes.
  //unsigned char temp;                 // used to convert bgr to rgb color.
  int i;

  // make sure the file is there.
  if ( ( file = fopen( fileName, "rb" ) ) == NULL ) 
  {
    gLog->add_msg( "File %s not Found", fileName );
    return false;
  }
  
  CTGAHeader header;
  fread( &header, sizeof( CTGAHeader ), 1, file );

  // We cannot have a colormap in this file
  if ( header.color_map_type != 0 )
  {
   gLog->add_msg( "Error: Targa image has colormap" );
    return false;
  }

  // Check image type
  if ( header.image_type != 2 )
  {
    gLog->add_msg( "Error: Targa image not true color" );
    return false;
  }

  tex->components = header.pixel_depth / 8;
  if ( tex->components != 4 )
  {
    gLog->add_msg( "Error: Targa image does not have 4 components per pixel" );
    return false;
  }

  // Skip past image ID bytes
  for ( i = 0; i < header.id_length; i++ )
    fgetc( file );

  // read the data.
  tex->sizeX = header.width;
  tex->sizeY = header.height;
  size = header.width * header.height * 4;
  tex->data = ( unsigned char * ) malloc( size );
  if ( tex->data == NULL ) 
  {
    gLog->add_msg( "Error allocating memory for color-corrected image data" );
    return 0;	
  }

  if ( fread( tex->data, size, 1, file ) != 1 ) 
  {
    gLog->add_msg( "Error reading image data from %s", fileName );
    return 0;
  }

  return true;
}

//////////////////////////////////////////////////////
bool
texture_manager_t::loadJPG( const char *fileName, CTexture *tex )
{
    JPEG_CORE_PROPERTIES image;
    ZeroMemory( &image, sizeof( JPEG_CORE_PROPERTIES ) );
    if( ijlInit( &image ) != IJL_OK )
	{
		gLog->add_msg( "Cannot initialize JLIB", fileName );
		return false;
	}

	image.JPGFile = fileName;
  if ( ijlRead( &image, IJL_JFILE_READPARAMS ) != IJL_OK )
  {
	  gLog->add_msg( "Cannot open %s", fileName );
		return false;
  }

    tex->sizeY = image.JPGHeight;
	tex->sizeX = image.JPGWidth;
	tex->components = 3;

	// read the data. 
	tex->data = ( unsigned char * ) malloc( tex->sizeX * tex->sizeY * tex->components );
    if ( !tex->data )
	{
		gLog->add_msg( "Cannot open %s", fileName );
		return false;
    }

    image.DIBBytes = tex->data;
    //image.DIBColor = IJL_RGB;
    image.DIBColor = IJL_BGR;
    image.DIBHeight = tex->sizeY;
    image.DIBWidth = tex->sizeX;
    image.DIBChannels = 3;
			
    if ( ijlRead( &image, IJL_JFILE_READWHOLEIMAGE ) != IJL_OK )
	{
		gLog->add_msg( "Cannot read %s", fileName );
		return false;
	}

    ijlFree( &image );
    return true;
}

//////////////////////////////////////////////////////////////////////
void 
texture_manager_t::saveBMP( const char *fileName, int numComponents, int w, int h,
		                    const unsigned char *data )
{
  FILE *file;
  if ( ( file = fopen( fileName, "wb" ) ) == NULL ) 
  {
    gLog->add_msg( "Cannot open file: ", fileName );
    return;
  }
  
  putshort( file, 0x4d42 );
  fseek( file, 8, SEEK_CUR );
  putint( file, 54 );
  putint( file, 0x28 );
  
  putint( file, w );
  putint( file, h );

  // write the planes
  putshort( file, 1 );
  
  // write the bpp
  putshort( file, numComponents * 8 );
  
  // seek past the rest of the bitmap header.
  fseek( file, 24, SEEK_CUR );

  unsigned char *tdata = new unsigned char [ w * h * numComponents ];
  for ( int i = 0; i < w * h * numComponents; i += 3 )
  {
    tdata[ i ] = data[ i + 2 ];
    tdata[ i + 1 ] = data[ i + 1 ];
    tdata[ i + 2 ] = data[ i ];
  }

  // write the data. 
  fwrite( tdata, 1, w * h * numComponents, file );
  delete [] tdata;
  fclose( file );
}
