#include "vect3.h"

/////////////////////////////////////////////////////////////////////////
// Distance from point to line segment
/////////////////////////////////////////////////////////////////////////
float 
segment_t::dist2_from_line_seg( const vect3_t &aP ) const
{
  static float lU;
  static vect3_t lIntersection;     
 
  lU = ( ( ( aP.x - mPoints[ 0 ].x ) * ( mP1MinusP0.x ) ) +
         ( ( aP.y - mPoints[ 0 ].y ) * ( mP1MinusP0.y ) ) +
         ( ( aP.z - mPoints[ 0 ].z ) * ( mP1MinusP0.z ) ) ) * mRecipMag2;
 
  if ( lU < 0.0f )
	{
		return aP.distanceSquared( mPoints[ 0 ] );	
	}
	else if ( lU > 1.0f )
	{
		return aP.distanceSquared( mPoints[ 1 ] );
	}
	else
	{
    lIntersection.x = mPoints[ 0 ].x + lU * ( mP1MinusP0.x );
	  lIntersection.y = mPoints[ 0 ].y + lU * ( mP1MinusP0.y );
	  lIntersection.z = mPoints[ 0 ].z + lU * ( mP1MinusP0.z );
 
	  return aP.distanceSquared( lIntersection );
	}
}

/////////////////////////////////////////////////
// Checks if the target point is inside the polygon
/////////////////////////////////////////////////
bool                                 
inpoly( const vector<vect3_t> &poly, /*   polygon points, [0]=x, [1]=y       */
        unsigned int xt,             /*   x (horizontal) of target point     */
        unsigned int yt )            /*   y (vertical) of target point       */
{
     unsigned int xnew,ynew;
     unsigned int xold,yold;
     unsigned int x1,y1;
     unsigned int x2,y2;
     int  i;
     bool inside = false;
     int  npoints = poly.size();

     if (npoints < 3) {
          return(0);
     }
     xold=poly[npoints-1].x;
     yold=poly[npoints-1].z;
     for (i=0 ; i < npoints; i++) {
          xnew=poly[i].x;
          ynew=poly[i].z;
          if (xnew > xold) {
               x1=xold;
               x2=xnew;
               y1=yold;
               y2=ynew;
          }
          else {
               x1=xnew;
               x2=xold;
               y1=ynew;
               y2=yold;
          }
          if ((xnew < xt) == (xt <= xold)           /* straddle test */
           && ((long)yt-(long)y1)*(long)(x2-x1)     /* north         */
            < ((long)y2-(long)y1)*(long)(xt-x1)) {  /*       test    */
               inside=!inside;
          }
          xold=xnew;
          yold=ynew;
     }
     return inside;
}

// Calculate the cross product and return it
static void cross (float dst[3], float srcA[3], float srcB[3])
{
    dst[0] = srcA[1]*srcB[2] - srcA[2]*srcB[1];
    dst[1] = srcA[2]*srcB[0] - srcA[0]*srcB[2];
    dst[2] = srcA[0]*srcB[1] - srcA[1]*srcB[0];
}

// Normalize the input vector
static void normalize (float vec[3])
{
    const float squaredLen = vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2];
    const float invLen = 1.f / (float) sqrt (squaredLen);

    vec[0] *= invLen;
    vec[1] *= invLen;
    vec[2] *= invLen;
}

// Scale the given vector
static void scale (float v[3], float s)
{
    v[0] *= s;
    v[1] *= s;
    v[2] *= s;
}

vect3_t gUp( 0, 1, 0 );

/////////////////////////////////////////////////////////////////////////
// Muliply a matrix onto the current one that look from aEye at aLookAt
/////////////////////////////////////////////////////////////////////////
void 
mult_look_at( const vect3_t &aEye,
		          const vect3_t &aLookat )
{
  static float m[ 16 ];
  float *xaxis = &m[ 0 ], *up = &m[ 4 ], *at = &m[ 8 ];

  // Compute our new look at vector, which will be
  //   the new negative Z axis of our transformed object.
  at[ 0 ] = aLookat.x - aEye.x; 
	at[ 1 ] = aLookat.y - aEye.y; 
	at[ 2 ] = aLookat.z - aEye.z;
  normalize( at );

  // Make a useable copy of the current up vector.
  up[ 0 ] = gUp.x; 
	up[ 1 ] = gUp.y; 
	up[ 2 ] = gUp.z;

  // Cross product of the new look at vector and the current
  //   up vector will produce a vector which is the new
  //   positive X axis of our transformed object.
  cross( xaxis, at, up );
  normalize( xaxis );

  // Calculate the new up vector, which will be the
  //   positive Y axis of our transformed object. Note
  //   that it will lie in the same plane as the new
  //   look at vector and the old up vector.
  cross( up, xaxis, at );

  // Account for the fact that the geometry will be defined to
  //   point along the negative Z axis.
  scale( at, -1.f );

  // Fill out the rest of the 4x4 matrix
  m[ 3 ] = 0.f;     // xaxis is m[0..2]
  m[ 7 ] = 0.f;     // up is m[4..6]
  m[ 11 ] = 0.f;    // -at is m[8..10]
  m[ 12 ] = aEye.x; 
  m[ 13 ] = aEye.y; 
  m[ 14 ] = aEye.z;
  m[ 15 ] = 1.f;

  // Multiply onto current matrix stack.
  glMultMatrixf( m );
}

void 
multLookAt( float eyex, float eyey, float eyez,
      	    float atx, float aty, float atz,
	          float upx, float upy, float upz )
{
    static float m[16];
    float *xaxis = &m[0],
        *up = &m[4],
        *at = &m[8];

    // Compute our new look at vector, which will be
    //   the new negative Z axis of our transformed object.
    at[0] = atx-eyex; at[1] = aty-eyey; at[2] = atz-eyez;
    normalize (at);

    // Make a useable copy of the current up vector.
    up[0] = upx; up[1] = upy; up[2] = upz;

    // Cross product of the new look at vector and the current
    //   up vector will produce a vector which is the new
    //   positive X axis of our transformed object.
    cross (xaxis, at, up);
    normalize (xaxis);

    // Calculate the new up vector, which will be the
    //   positive Y axis of our transformed object. Note
    //   that it will lie in the same plane as the new
    //   look at vector and the old up vector.
    cross (up, xaxis, at);

    // Account for the fact that the geometry will be defined to
    //   point along the negative Z axis.
    scale (at, -1.f);

    // Fill out the rest of the 4x4 matrix
    m[3] = 0.f;     // xaxis is m[0..2]
    m[7] = 0.f;     // up is m[4..6]
    m[11] = 0.f;    // -at is m[8..10]
    m[12] = eyex; m[13] = eyey; m[14] = eyez;
    m[15] = 1.f;

    // Multiply onto current matrix stack.
    glMultMatrixf (m);
}

////////////////////////////////////////////////////////////////
// v is normalized
// theta in radians
////////////////////////////////////////////////////////////////
void 
mat4_t::set_rot( float theta, const vect3_t &v ) 
{
    float ct = float( cos( theta ) );
    float st = float( sin( theta ) );

    float xx = v.x * v.x;
    float yy = v.y * v.y;
    float zz = v.z * v.z;
    float xy = v.x * v.y;
    float xz = v.x * v.z;
    float yz = v.y * v.z;

    a00 = xx + ct*(1-xx);
    a01 = xy + ct*(-xy) + st*-v.z;
    a02 = xz + ct*(-xz) + st*v.y;

    a10 = xy + ct*(-xy) + st*v.z;
    a11 = yy + ct*(1-yy);
    a12 = yz + ct*(-yz) + st*-v.x;

    a20 = xz + ct*(-xz) + st*-v.y;
    a21 = yz + ct*(-yz) + st*v.x;
    a22 = zz + ct*(1-zz);
}